﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.EvolutionTracking {
  [StorableClass]
  [Item("AfterCrossoverOperator", "Performs an action after the crossover operator is applied.")]
  public class BeforeManipulatorOperator<TGraph, TVertex, TContent> : EvolutionTrackingOperator<TGraph, TVertex, TContent>
    where TGraph : class,IGenealogyGraph<TVertex, TContent>
    where TVertex : class,IGenealogyGraphNode<TContent>, new()
    where TContent : class,IItem {

    private const string ChildParameterName = "Child";
    public ILookupParameter<TContent> ChildParameter;

    protected BeforeManipulatorOperator(BeforeManipulatorOperator<TGraph, TVertex, TContent> original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new BeforeManipulatorOperator<TGraph, TVertex, TContent>(this, cloner);
    }

    public BeforeManipulatorOperator() {
      ChildParameter = new LookupParameter<TContent>(ChildParameterName);
      Parameters.Add(ChildParameter);
    }

    public override IOperation Apply() {
      if (GenealogyGraph.Contains(ChildParameter.ActualValue)) {
        // if the individual has been affected by crossover before mutation, we keep it in the graph and feed a clone of it to the mutation operator
        var child = ChildParameter.ActualValue;
        var clone = (TContent)child.Clone();
        var vertex = GenealogyGraph[child].Last();
        vertex.Content = clone;
        GenealogyGraph.Ranks[vertex.Rank].Remove(vertex);
        var newVertex = new TVertex {
          Content = child,
          Rank = vertex.Rank
        };
        vertex.Rank -= 0.5;
        if (!GenealogyGraph.Ranks.ContainsKey(vertex.Rank)) {
          GenealogyGraph.Ranks[vertex.Rank] = new LinkedList<TVertex>();
        }
        GenealogyGraph.Ranks[vertex.Rank].AddLast(vertex);

        newVertex.AddForwardArc(vertex);
        vertex.AddReverseArc(vertex);
      } else {
        // mutation should never happen without crossover 
      }
      return base.Apply();
    }
  }
}
